@@ -7,7 +7,7 @@
  * ------------
  * Description:
  * ------------
- * This is an implemention of Unicode's Bidirectional Algorithm
+ * This is an implementation of Unicode's Bidirectional Algorithm
  * (known as UAX #9).
  *
  *   http://www.unicode.org/reports/tr9/
@@ -17,9 +17,22 @@
  *
  * Modified: Thomas Wolff:
  * - extended bidi class detection to consider non-BMP characters
- * - fixed mirroring of '('
+ * - generating mirroring data from Unicode
  * - fixed recognition of runs
  * - sanitized handling of directional markers
+ * - support bidi mode model of «BiDi in Terminal Emulators» recommendation
+     (https://terminal-wg.pages.freedesktop.org/bidi/):
+ *   - return resolved paragraph level (or -1 if none)
+ *   - switchable options for:
+ *     - autodetection (rules P2, P3)
+ *     - LTR/RTL fallback for autodetection
+ *     - explicit RTL
+ *     - mirroring of additional characters (Box drawing, quadrant blocks)
+ * - fixed UBA shortcut which ignored AN
+ * - implemented UBA rule N0 (bracket pair handling)
+ * - fixed X9 to mask formatters with NSM rather than BN
+ * - fixed W7 to fallback to sor
+ * - fixed L1 to skip directional markers
  *
  ************************************************************************/
 
@@ -337,91 +350,71 @@ do_shape(bidi_char * line, bidi_char * to, int count)
   return 1;
 }
 
+enum { BRACKx, BRACKo, BRACKc };
+
 static ucschar
-mirror(ucschar c)
+bracket(ucschar c)
+{
+  static const struct {
+    wchar from, to;
+    uchar bracket;
+  } pairs[] = {
+#include "brackets.t"
+  };
+
+  int i = -1;
+  int j = lengthof(pairs);
+  while (j - i > 1) {
+    int k = (i + j) / 2;
+    if (c == pairs[k].from) {
+      if (pairs[k].bracket == BRACKo)
+        return c;
+      else
+        return pairs[k].to;
+    }
+    if (c < pairs[k].from)
+      j = k;
+    else
+      i = k;
+  }
+
+  return 0;
+}
+
+/*
+  Determine canonical equivalent of a bracket character.
+ */
+static ucschar
+canonical(ucschar c)
+{
+  static const struct {
+    wchar bracket, canonical;
+  } pairs[] = {
+#include "canonical.t"
+  };
+
+  int i = -1;
+  int j = lengthof(pairs);
+  while (j - i > 1) {
+    int k = (i + j) / 2;
+    if (c == pairs[k].bracket)
+      return pairs[k].canonical;
+    if (c < pairs[k].bracket)
+      j = k;
+    else
+      i = k;
+  }
+
+  return c;
+}
+
+static ucschar
+mirror(ucschar c, bool box_mirror)
 {
   static const struct { wchar from, to; } pairs[] = {
-    {0x0028, 0x0029}, {0x0029, 0x0028}, {0x003C, 0x003E}, {0x003E, 0x003C},
-    {0x005B, 0x005D}, {0x005D, 0x005B}, {0x007B, 0x007D}, {0x007D, 0x007B},
-    {0x00AB, 0x00BB}, {0x00BB, 0x00AB}, {0x2039, 0x203A}, {0x203A, 0x2039},
-    {0x2045, 0x2046}, {0x2046, 0x2045}, {0x207D, 0x207E}, {0x207E, 0x207D},
-    {0x208D, 0x208E}, {0x208E, 0x208D}, {0x2208, 0x220B}, {0x2209, 0x220C},
-    {0x220A, 0x220D}, {0x220B, 0x2208}, {0x220C, 0x2209}, {0x220D, 0x220A},
-    {0x2215, 0x29F5}, {0x223C, 0x223D}, {0x223D, 0x223C}, {0x2243, 0x22CD},
-    {0x2252, 0x2253}, {0x2253, 0x2252}, {0x2254, 0x2255}, {0x2255, 0x2254},
-    {0x2264, 0x2265}, {0x2265, 0x2264}, {0x2266, 0x2267}, {0x2267, 0x2266},
-    {0x2268, 0x2269}, {0x2269, 0x2268}, {0x226A, 0x226B}, {0x226B, 0x226A},
-    {0x226E, 0x226F}, {0x226F, 0x226E}, {0x2270, 0x2271}, {0x2271, 0x2270},
-    {0x2272, 0x2273}, {0x2273, 0x2272}, {0x2274, 0x2275}, {0x2275, 0x2274},
-    {0x2276, 0x2277}, {0x2277, 0x2276}, {0x2278, 0x2279}, {0x2279, 0x2278},
-    {0x227A, 0x227B}, {0x227B, 0x227A}, {0x227C, 0x227D}, {0x227D, 0x227C},
-    {0x227E, 0x227F}, {0x227F, 0x227E}, {0x2280, 0x2281}, {0x2281, 0x2280},
-    {0x2282, 0x2283}, {0x2283, 0x2282}, {0x2284, 0x2285}, {0x2285, 0x2284},
-    {0x2286, 0x2287}, {0x2287, 0x2286}, {0x2288, 0x2289}, {0x2289, 0x2288},
-    {0x228A, 0x228B}, {0x228B, 0x228A}, {0x228F, 0x2290}, {0x2290, 0x228F},
-    {0x2291, 0x2292}, {0x2292, 0x2291}, {0x2298, 0x29B8}, {0x22A2, 0x22A3},
-    {0x22A3, 0x22A2}, {0x22A6, 0x2ADE}, {0x22A8, 0x2AE4}, {0x22A9, 0x2AE3},
-    {0x22AB, 0x2AE5}, {0x22B0, 0x22B1}, {0x22B1, 0x22B0}, {0x22B2, 0x22B3},
-    {0x22B3, 0x22B2}, {0x22B4, 0x22B5}, {0x22B5, 0x22B4}, {0x22B6, 0x22B7},
-    {0x22B7, 0x22B6}, {0x22C9, 0x22CA}, {0x22CA, 0x22C9}, {0x22CB, 0x22CC},
-    {0x22CC, 0x22CB}, {0x22CD, 0x2243}, {0x22D0, 0x22D1}, {0x22D1, 0x22D0},
-    {0x22D6, 0x22D7}, {0x22D7, 0x22D6}, {0x22D8, 0x22D9}, {0x22D9, 0x22D8},
-    {0x22DA, 0x22DB}, {0x22DB, 0x22DA}, {0x22DC, 0x22DD}, {0x22DD, 0x22DC},
-    {0x22DE, 0x22DF}, {0x22DF, 0x22DE}, {0x22E0, 0x22E1}, {0x22E1, 0x22E0},
-    {0x22E2, 0x22E3}, {0x22E3, 0x22E2}, {0x22E4, 0x22E5}, {0x22E5, 0x22E4},
-    {0x22E6, 0x22E7}, {0x22E7, 0x22E6}, {0x22E8, 0x22E9}, {0x22E9, 0x22E8},
-    {0x22EA, 0x22EB}, {0x22EB, 0x22EA}, {0x22EC, 0x22ED}, {0x22ED, 0x22EC},
-    {0x22F0, 0x22F1}, {0x22F1, 0x22F0}, {0x22F2, 0x22FA}, {0x22F3, 0x22FB},
-    {0x22F4, 0x22FC}, {0x22F6, 0x22FD}, {0x22F7, 0x22FE}, {0x22FA, 0x22F2},
-    {0x22FB, 0x22F3}, {0x22FC, 0x22F4}, {0x22FD, 0x22F6}, {0x22FE, 0x22F7},
-    {0x2308, 0x2309}, {0x2309, 0x2308}, {0x230A, 0x230B}, {0x230B, 0x230A},
-    {0x2329, 0x232A}, {0x232A, 0x2329}, {0x2768, 0x2769}, {0x2769, 0x2768},
-    {0x276A, 0x276B}, {0x276B, 0x276A}, {0x276C, 0x276D}, {0x276D, 0x276C},
-    {0x276E, 0x276F}, {0x276F, 0x276E}, {0x2770, 0x2771}, {0x2771, 0x2770},
-    {0x2772, 0x2773}, {0x2773, 0x2772}, {0x2774, 0x2775}, {0x2775, 0x2774},
-    {0x27D5, 0x27D6}, {0x27D6, 0x27D5}, {0x27DD, 0x27DE}, {0x27DE, 0x27DD},
-    {0x27E2, 0x27E3}, {0x27E3, 0x27E2}, {0x27E4, 0x27E5}, {0x27E5, 0x27E4},
-    {0x27E6, 0x27E7}, {0x27E7, 0x27E6}, {0x27E8, 0x27E9}, {0x27E9, 0x27E8},
-    {0x27EA, 0x27EB}, {0x27EB, 0x27EA}, {0x2983, 0x2984}, {0x2984, 0x2983},
-    {0x2985, 0x2986}, {0x2986, 0x2985}, {0x2987, 0x2988}, {0x2988, 0x2987},
-    {0x2989, 0x298A}, {0x298A, 0x2989}, {0x298B, 0x298C}, {0x298C, 0x298B},
-    {0x298D, 0x2990}, {0x298E, 0x298F}, {0x298F, 0x298E}, {0x2990, 0x298D},
-    {0x2991, 0x2992}, {0x2992, 0x2991}, {0x2993, 0x2994}, {0x2994, 0x2993},
-    {0x2995, 0x2996}, {0x2996, 0x2995}, {0x2997, 0x2998}, {0x2998, 0x2997},
-    {0x29B8, 0x2298}, {0x29C0, 0x29C1}, {0x29C1, 0x29C0}, {0x29C4, 0x29C5},
-    {0x29C5, 0x29C4}, {0x29CF, 0x29D0}, {0x29D0, 0x29CF}, {0x29D1, 0x29D2},
-    {0x29D2, 0x29D1}, {0x29D4, 0x29D5}, {0x29D5, 0x29D4}, {0x29D8, 0x29D9},
-    {0x29D9, 0x29D8}, {0x29DA, 0x29DB}, {0x29DB, 0x29DA}, {0x29F5, 0x2215},
-    {0x29F8, 0x29F9}, {0x29F9, 0x29F8}, {0x29FC, 0x29FD}, {0x29FD, 0x29FC},
-    {0x2A2B, 0x2A2C}, {0x2A2C, 0x2A2B}, {0x2A2D, 0x2A2C}, {0x2A2E, 0x2A2D},
-    {0x2A34, 0x2A35}, {0x2A35, 0x2A34}, {0x2A3C, 0x2A3D}, {0x2A3D, 0x2A3C},
-    {0x2A64, 0x2A65}, {0x2A65, 0x2A64}, {0x2A79, 0x2A7A}, {0x2A7A, 0x2A79},
-    {0x2A7D, 0x2A7E}, {0x2A7E, 0x2A7D}, {0x2A7F, 0x2A80}, {0x2A80, 0x2A7F},
-    {0x2A81, 0x2A82}, {0x2A82, 0x2A81}, {0x2A83, 0x2A84}, {0x2A84, 0x2A83},
-    {0x2A8B, 0x2A8C}, {0x2A8C, 0x2A8B}, {0x2A91, 0x2A92}, {0x2A92, 0x2A91},
-    {0x2A93, 0x2A94}, {0x2A94, 0x2A93}, {0x2A95, 0x2A96}, {0x2A96, 0x2A95},
-    {0x2A97, 0x2A98}, {0x2A98, 0x2A97}, {0x2A99, 0x2A9A}, {0x2A9A, 0x2A99},
-    {0x2A9B, 0x2A9C}, {0x2A9C, 0x2A9B}, {0x2AA1, 0x2AA2}, {0x2AA2, 0x2AA1},
-    {0x2AA6, 0x2AA7}, {0x2AA7, 0x2AA6}, {0x2AA8, 0x2AA9}, {0x2AA9, 0x2AA8},
-    {0x2AAA, 0x2AAB}, {0x2AAB, 0x2AAA}, {0x2AAC, 0x2AAD}, {0x2AAD, 0x2AAC},
-    {0x2AAF, 0x2AB0}, {0x2AB0, 0x2AAF}, {0x2AB3, 0x2AB4}, {0x2AB4, 0x2AB3},
-    {0x2ABB, 0x2ABC}, {0x2ABC, 0x2ABB}, {0x2ABD, 0x2ABE}, {0x2ABE, 0x2ABD},
-    {0x2ABF, 0x2AC0}, {0x2AC0, 0x2ABF}, {0x2AC1, 0x2AC2}, {0x2AC2, 0x2AC1},
-    {0x2AC3, 0x2AC4}, {0x2AC4, 0x2AC3}, {0x2AC5, 0x2AC6}, {0x2AC6, 0x2AC5},
-    {0x2ACD, 0x2ACE}, {0x2ACE, 0x2ACD}, {0x2ACF, 0x2AD0}, {0x2AD0, 0x2ACF},
-    {0x2AD1, 0x2AD2}, {0x2AD2, 0x2AD1}, {0x2AD3, 0x2AD4}, {0x2AD4, 0x2AD3},
-    {0x2AD5, 0x2AD6}, {0x2AD6, 0x2AD5}, {0x2ADE, 0x22A6}, {0x2AE3, 0x22A9},
-    {0x2AE4, 0x22A8}, {0x2AE5, 0x22AB}, {0x2AEC, 0x2AED}, {0x2AED, 0x2AEC},
-    {0x2AF7, 0x2AF8}, {0x2AF8, 0x2AF7}, {0x2AF9, 0x2AFA}, {0x2AFA, 0x2AF9},
-    {0x3008, 0x3009}, {0x3009, 0x3008}, {0x300A, 0x300B}, {0x300B, 0x300A},
-    {0x300C, 0x300D}, {0x300D, 0x300C}, {0x300E, 0x300F}, {0x300F, 0x300E},
-    {0x3010, 0x3011}, {0x3011, 0x3010}, {0x3014, 0x3015}, {0x3015, 0x3014},
-    {0x3016, 0x3017}, {0x3017, 0x3016}, {0x3018, 0x3019}, {0x3019, 0x3018},
-    {0x301A, 0x301B}, {0x301B, 0x301A}, {0xFF08, 0xFF09}, {0xFF09, 0xFF08},
-    {0xFF1C, 0xFF1E}, {0xFF1E, 0xFF1C}, {0xFF3B, 0xFF3D}, {0xFF3D, 0xFF3B},
-    {0xFF5B, 0xFF5D}, {0xFF5D, 0xFF5B}, {0xFF5F, 0xFF60}, {0xFF60, 0xFF5F},
-    {0xFF62, 0xFF63}, {0xFF63, 0xFF62}
+#include "mirroring.t"
   };
+
   int i = -1;
   int j = lengthof(pairs);
   while (j - i > 1) {
@@ -433,9 +426,70 @@ mirror(ucschar c)
     else
       i = k;
   }
+
+  /* check Box Drawing (U+2500-U+257F) and Block Elements (U+2580-U+259F)
+  ┌ ┍ ┎ ┏ └ ┕ ┖ ┗ ├ ┝ ┞ ┟ ┠ ┡ ┢ ┣ ┭ ┱ ┵ ┹ ┽ ╃ ╅ ╉
+  ┐ ┑ ┒ ┓ ┘ ┙ ┚ ┛ ┤ ┥ ┦ ┧ ┨ ┩ ┪ ┫ ┮ ┲ ┶ ┺ ┾ ╄ ╆ ╊
+  ╒ ╓ ╔ ╘ ╙ ╚ ╞ ╟ ╠ ╭ ╯ ╱ ╴ ╸ ╼
+  ╕ ╖ ╗ ╛ ╜ ╝ ╡ ╢ ╣ ╮ ╰ ╲ ╶ ╺ ╾
+
+  cannot handle quarter/eighth blocks (would need to be reversed too) ▉▊▋▍▎
+  could handle one eighth blocks ▏▕ but would be inconsistent
+
+  ▌ ▖ ▘ ▙ ▚ ▛
+  ▐ ▗ ▝ ▟ ▞ ▜
+  */
+  static const struct { wchar from, to; } boxpairs[] = {
+    {0x250C, 0x2510}, {0x250D, 0x2511}, {0x250E, 0x2512}, {0x250F, 0x2513},
+    {0x2510, 0x250C}, {0x2511, 0x250D}, {0x2512, 0x250E}, {0x2513, 0x250F},
+    {0x2514, 0x2518}, {0x2515, 0x2519}, {0x2516, 0x251A}, {0x2517, 0x251B},
+    {0x2518, 0x2514}, {0x2519, 0x2515}, {0x251A, 0x2516}, {0x251B, 0x2517},
+    {0x251C, 0x2524}, {0x251D, 0x2525}, {0x251E, 0x2526}, {0x251F, 0x2527},
+    {0x2520, 0x2528}, {0x2521, 0x2529}, {0x2522, 0x252A}, {0x2523, 0x252B},
+    {0x2524, 0x251C}, {0x2525, 0x251D}, {0x2526, 0x251E}, {0x2527, 0x251F},
+    {0x2528, 0x2520}, {0x2529, 0x2521}, {0x252A, 0x2522}, {0x252B, 0x2523},
+    {0x252D, 0x252E}, {0x252E, 0x252D}, {0x2531, 0x2532}, {0x2532, 0x2531},
+    {0x2535, 0x2536}, {0x2536, 0x2535}, {0x2539, 0x253A}, {0x253A, 0x2539},
+    {0x253D, 0x253E}, {0x253E, 0x253D}, {0x2543, 0x2544}, {0x2544, 0x2543},
+    {0x2545, 0x2546}, {0x2546, 0x2545}, {0x2549, 0x254A}, {0x254A, 0x2549},
+    {0x2552, 0x2555}, {0x2553, 0x2556}, {0x2554, 0x2557}, {0x2555, 0x2552},
+    {0x2556, 0x2553}, {0x2557, 0x2554}, {0x2558, 0x255B}, {0x2559, 0x255C},
+    {0x255A, 0x255D}, {0x255B, 0x2558}, {0x255C, 0x2559}, {0x255D, 0x255A},
+    {0x255E, 0x2561}, {0x255F, 0x2562}, {0x2560, 0x2563}, {0x2561, 0x255E},
+    {0x2562, 0x255F}, {0x2563, 0x2560}, {0x256D, 0x256E}, {0x256E, 0x256D},
+    {0x256F, 0x2570}, {0x2570, 0x256F}, {0x2571, 0x2572}, {0x2572, 0x2571},
+    {0x2574, 0x2576}, {0x2576, 0x2574}, {0x2578, 0x257A}, {0x257A, 0x2578},
+    {0x257C, 0x257E}, {0x257E, 0x257C}, {0x258C, 0x2590}, {0x2590, 0x258C},
+    {0x2596, 0x2597}, {0x2597, 0x2596}, {0x2598, 0x259D}, {0x2599, 0x259F},
+    {0x259A, 0x259E}, {0x259B, 0x259C}, {0x259C, 0x259B}, {0x259D, 0x2598},
+    {0x259E, 0x259A}, {0x259F, 0x2599},
+  };
+
+  if (!box_mirror)
+    return c;
+
+  i = -1;
+  j = lengthof(boxpairs);
+  while (j - i > 1) {
+    int k = (i + j) / 2;
+    if (c == boxpairs[k].from)
+      return boxpairs[k].to;
+    if (c < boxpairs[k].from)
+      j = k;
+    else
+      i = k;
+  }
+
   return c;
 }
 
+
+#ifdef TEST_BIDI
+uchar bidi_levels[999];
+# define debug_bidi
+int do_trace_bidi = 0;
+#endif
+
 /*
  * The Main Bidi Function, and the only function that should
  * be used by the outside world.
@@ -443,40 +497,49 @@ mirror(ucschar c)
  * line: a buffer of size count containing text to apply
  * the Bidirectional algorithm to.
  */
-
 int
-do_bidi(bidi_char * line, int count)
+do_bidi(bool autodir, int paragraphLevel, bool explicitRTL, bool box_mirror, 
+        bidi_char * line, int count)
 {
-  uchar paragraphLevel;
   uchar currentEmbedding;
   uchar currentOverride;
   uchar tempType;
-  int i, j, yes;
+  int i, j;
 
   uchar bidi_class_of(int i) {
-    if (i && line[i].wc == UCSWIDE)
-      i--;
-    return bidi_class(line[i].wc);
-  }
+    ucschar c = line[i].wc;
 
- /* Check the presence of R or AL types as optimization */
-  yes = 0;
-  for (i = 0; i < count; i++) {
-    int type = bidi_class_of(i);
-    if (type == R || type == AL
-        || type == RLE || type == LRE || type == RLO || type == LRO || type == PDF
-        || type == LRI || type == RLI || type == FSI || type == PDI
-       ) {
-      yes = 1;
-      break;
+#ifdef try_to_handle_CJK_here
+#warning does not always work in RTL mode, now filtered before calling do_bidi
+    if (i && c == UCSWIDE) {
+      // try to fix double-width within right-to-left
+      if (currentEmbedding & 1)
+        i--;
+      else
+        return BN;
+      // OK for LTR: return BN
+      // OK for RTL U+5555: EN, NSM
+      // not displayed in RTL: U+FF1C (class default -> ON)
     }
-  }
-  if (yes == 0)
-    return L;
+#endif
 
- /* Initialize types, levels */
-  uchar types[count];
-  uchar levels[count];
+#ifdef check_emoji
+    if (c == 0x200D
+     || (c >= 0x2600 && c < 0x2800)
+     || (c >= 0x1F300 && c < 0x20000)
+       )
+     return EN;
+#endif
+#ifdef check_emojilen
+    if (line[i].emojilen)
+      return EN;
+#endif
+
+    if (explicitRTL)
+      return R;
+
+    return bidi_class(c);
+  }
 
  /* Rule (P1)  NOT IMPLEMENTED
   * P1. Split the text into separate paragraphs. A paragraph separator is
@@ -491,23 +554,122 @@ do_bidi(bidi_char * line, int count)
   * P3. If a character is found in P2 and it is of type AL or R, then set
   * the paragraph embedding level to one; otherwise, set it to zero.
   */
-  paragraphLevel = 0;
   int isolateLevel = 0;
+  int resLevel = -1;
+  bool hasRTL = false;
   for (i = 0; i < count; i++) {
     int type = bidi_class_of(i);
-    if (type == LRI || type == RLI || type == FSI)
+    if (type == LRI || type == RLI || type == FSI) {
+      hasRTL = true;
       isolateLevel++;
-    else if (type == PDI)
+    }
+    else if (type == PDI) {
+      hasRTL = true;
       isolateLevel--;
+    }
     else if (isolateLevel == 0) {
       if (type == R || type == AL) {
-        paragraphLevel = 1;
+        hasRTL = true;
+        if (resLevel < 0)
+          resLevel = 1;
         break;
       }
-      else if (type == L)
-        break;
+      else if (type == RLE || type == LRE || type == RLO || type == LRO || type == PDF) {
+        hasRTL = true;
+        if (resLevel >= 0)
+          break;
+      }
+      else if (type == L) {
+        if (resLevel < 0)
+          resLevel = 0;
+      }
+      else if (type == AN)
+        hasRTL = true;
     }
   }
+  if (autodir) {
+    if (resLevel >= 0)
+      paragraphLevel = resLevel;
+  }
+  else
+    resLevel = paragraphLevel;
+ /* Optimization: skip full algorithm if there is nothing to reorder */
+  if (!hasRTL && !paragraphLevel)
+    return 0;
+
+
+ /* Initialize types, levels */
+  uchar types[count];
+  uchar levels[count];
+  // workaround for gcc 11 warning anomaly
+  types[0] = 0;
+
+#define dont_debug_bidi
+
+#ifdef debug_bidi
+  uchar prev_types[count];
+  uchar prev_levels[count];
+
+  void trace_bidi(char * tag)
+  {
+#ifndef TEST_BIDI
+    static int do_trace_bidi = 2;
+#endif
+    if (do_trace_bidi) {
+      if (!tag) {
+        do_trace_bidi--;
+        return;
+      }
+
+      uint last = count - 1;
+      while (last && line[last].wc == ' ')
+        last--;
+
+      bool vacuous = true;
+      for (uint i = 0; i <= last; i++)
+        if (types[i] != prev_types[i] || levels[i] != prev_levels[i]) {
+          vacuous = false;
+          break;
+        }
+      for (uint i = 0; i <= last; i++) {
+        prev_types[i] = types[i];
+        prev_levels[i] = levels[i];
+      }
+      if (vacuous)
+        return;
+
+      printf("%s\n", tag);
+      for (uint i = 0; i <= last; i++)
+        printf(" %04X", line[i].wc);
+      printf("\n");
+      static char * _type[] = {
+        "  L", "LRE", "LRO", "  R", " AL", "RLE", "RLO", "PDF", " EN", " ES", " ET", " AN", " CS", "NSM", " BN", "  B", "  S", " WS", " ON", "LRI", "RLI", "FSI", "PDI"
+      };
+      for (uint i = 0; i <= last; i++)
+        if (types[i] < lengthof(_type))
+          printf("  %s", _type[types[i]]);
+        else
+          printf("    ?");
+      printf("\n");
+      for (uint i = 0; i <= last; i++)
+          printf(" %4d", levels[i]);
+      printf("\n");
+    }
+  }
+  void trace_mark(char * tag) {
+    (void)tag;
+  }
+  trace_bidi(0);
+  for (i = 0; i < count; i++) {
+    types[i] = ON;
+    levels[i] = -1;
+  }
+#else
+#define trace_bidi(tag)	
+#define trace_mark(tag)	
+#endif
+
+  trace_bidi("[P2, P3]");
 
  /* Rule (X1)
     X1. At the beginning of a paragraph, perform the following steps:
@@ -558,31 +720,6 @@ do_bidi(bidi_char * line, int count)
   //int ovfEmbedding = 0;
   isolateLevel = 0;
 
-#define dont_debug_bidi
-
-#ifdef debug_bidi
-  void trace_bidi(char * tag) {
-    if ((*line).wc == 'x') {
-      printf("%s ", tag);
-      for (int i = 0; i < count; i++) {
-        if (line[i].wc == ' ')
-          printf(" ");
-        else if (line[i].wc < 0x80)
-          printf("%d:%c/%d ", levels[i], line[i].wc, types[i]);
-        else
-          printf("%d:%X/%d ", levels[i], line[i].wc, types[i]);
-      }
-      printf("\n");
-    }
-  }
-  void trace_mark(char * tag) {
-    (void)tag;
-  }
-#else
-#define trace_bidi(tag)	
-#define trace_mark(tag)	
-#endif
-
  /* Rule (X2), (X3), (X4), (X5), (X6), (X7), (X8)
   * X2. With each RLE, compute the least greater odd embedding level.
   * X3. With each LRE, compute the least greater even embedding level.
@@ -686,7 +823,6 @@ do_bidi(bidi_char * line, int count)
         levels[i] = currentEmbedding;
         trace_mark("PDI");
       when WS or S: /* Whitespace is treated as neutral for now */
-        tempType = ON;
         if (currentOverride != ON)
           tempType = currentOverride;
       otherwise:
@@ -695,18 +831,24 @@ do_bidi(bidi_char * line, int count)
     }
     types[i] = tempType;
   }
+  trace_bidi("[X1-X8]");
 
-  trace_bidi("<X9");
  /* Rule (X9)
   * X9. Remove all RLE, LRE, RLO, LRO, PDF, and BN codes.
-  * Here, they're converted to BN.
+  * Here, they're converted to NSM (used to be BN).
   */
+  bool skip[count];
   for (i = 0; i < count; i++) {
     switch (types[i]) {
-      when RLE or LRE or RLO or LRO or PDF:
-        types[i] = BN;
+      when RLE or LRE or RLO or LRO or PDF or BN:
+        //types[i] = BN;
+        types[i] = NSM;  // fixes 4594 test cases + 28 char test cases
+        skip[i] = true;  // remove char from algorithm... (usage incomplete)
+      otherwise:
+        skip[i] = false;
     }
   }
+  trace_bidi("[X9]");
 
  /* Rule (X10) NOT IMPLEMENTED
   * X10. Handle isolating run sequences...
@@ -726,11 +868,11 @@ do_bidi(bidi_char * line, int count)
     (Note that in an isolating run sequence, an isolate initiator followed by 
     an NSM or any type other than PDI must be an overflow isolate initiator.)
   */
-  if (types[0] == NSM)
-    types[0] = paragraphLevel;
+  if (types[0] == NSM /*&& !skip[0]*/)
+    types[0] = (paragraphLevel & 1) ? R : L;  // sor
 
   for (i = 1; i < count; i++) {
-    if (types[i] == NSM)
+    if (types[i] == NSM /*&& !skip[i]*/)
       switch (types[i - 1]) {
         when LRI or RLI or FSI or PDI:
           types[i] = ON;
@@ -738,6 +880,7 @@ do_bidi(bidi_char * line, int count)
           types[i] = types[i - 1];
       }
   }
+  trace_bidi("[W1]");
 
  /* Rule (W2)
   * W2. Search backwards from each instance of a European number until the
@@ -759,6 +902,7 @@ do_bidi(bidi_char * line, int count)
       }
     }
   }
+  trace_bidi("[W2]");
 
  /* Rule (W3)
   * W3. Change all ALs to R.
@@ -770,13 +914,14 @@ do_bidi(bidi_char * line, int count)
     if (types[i] == AL)
       types[i] = R;
   }
+  trace_bidi("[W3]");
 
  /* Rule (W4)
   * W4. A single European separator between two European numbers changes
   * to a European number. A single common separator between two numbers
   * of the same type changes to that type.
   */
-  for (i = 1; i < (count - 1); i++) {
+  for (i = 1; i < count - 1; i++) {
     if (types[i] == ES) {
       if (types[i - 1] == EN && types[i + 1] == EN)
         types[i] = EN;
@@ -825,15 +970,70 @@ do_bidi(bidi_char * line, int count)
         types[i] = ON;
     }
   }
+  trace_bidi("[W4] [W5] [W6]");
+
+#define dont_consider_BD13
+
+#ifdef consider_BD13
+  int isol_run_level;
+#endif
+  void clear_isol()
+  {
+#ifdef consider_BD13
+    isol_run_level = 0;
+#endif
+  }
+  bool break_isol(int j)
+  {
+    if (!j)
+      return true;
+#ifdef consider_BD13
+    // BD13 describes "isolating run sequences" for use according to X10;
+    // however, it defines only isolate marks (particularly PDI) as 
+    // boundaries, inconsistently with some examples in BD13 and X10;
+    // this whole attempt does not seem to make much difference anyway...
+    // (-38 +18 BidiTest cases, all irregular / unsymmetric markers)
+    if (types[j] == PDI) {
+      isol_run_level++;
+      return false;
+    }
+    else if (bidi_class_of(j) == RLI || bidi_class_of(j) == LRI || bidi_class_of(j) == FSI) {
+      if (isol_run_level)
+        isol_run_level--;
+      return false;
+    }
+    else if (isol_run_level)
+      return false;
+# ifdef consider_X10_Example_1
+    // does not make a difference
+    else if (bidi_class_of(j - 1) == PDF && (bidi_class_of(j) == RLE || bidi_class_of(j) == LRE)) {
+      // X10 Example 1
+      return false;
+    }
+# endif
+    else {
+      j--;
+      if (bidi_class_of(j) == RLE || bidi_class_of(j) == LRE)
+        return true;
+# ifdef break_at_PDF
+      if (bidi_class_of(j) == PDF)
+        return true;
+# endif
+    }
+#endif
+    return false;
+  }
 
  /* Rule (W7)
   * W7. Search backwards from each instance of a European number until
   * the first strong type (R, L, or sor) is found. If an L is found,
   * then change the type of the European number to L.
   */
+  int embeddingLevel = paragraphLevel;  //TODO: sor
   for (i = 0; i < count; i++) {
     if (types[i] == EN) {
       j = i;
+      clear_isol();
       while (j >= 0) {
         if (types[j] == L) {
           types[i] = L;
@@ -842,14 +1042,140 @@ do_bidi(bidi_char * line, int count)
         else if (types[j] == R || types[j] == AL) {
           break;
         }
+        if (break_isol(j)) {
+          // nothing found, fallback to sor
+          if (!(embeddingLevel & 1))
+            types[i] = L;
+          break;
+        }
         j--;
       }
     }
   }
+  trace_bidi("[W7]");
 
- /* Rule (N0) NOT IMPLEMENTED
-  * N0. Handle bracket pairs in isolating run sequences...
+ /* Rule (N0) Handle bracket pairs in isolating run sequences.
+  * N0. Process bracket pairs in an isolating run sequence 
+    sequentially in the logical order of the text positions of the 
+    opening paired brackets using the logic given below.
+    Within this scope, bidirectional types EN and AN are treated as R.
   */
+#define is_N0_R(type) (type == R || type == EN || type == AN)
+  // refer to embedding level
+  embeddingLevel = paragraphLevel;  //TODO: to become sos when we handle isolating runs
+  // create stack and list of bracket pairs
+  struct {
+    ucschar b;
+    int o;
+  } brackstack[count];
+  int top = 0;
+  struct {
+    int o, c;
+  } brackpairs[count];
+  int bracks = 0;
+  // look for bracket pairs
+  // • Identify the bracket pairs in the current isolating run sequence.
+  for (i = 0; i < count; i++) {
+    ucschar bro = bracket(line[i].wc);
+    if (bro == line[i].wc) {  // opening bracket
+      brackstack[top].b = canonical(bro);
+      brackstack[top].o = i;
+      //printf("N0 > bracket %d ...\n", i);
+      top++;
+    }
+    else if (bro) {  // closing bracket, bro is the opening bracket
+      int sp = top;
+      while (sp) {
+        sp--;
+        // compare canonical equivalents, factoring out 
+        // SUPERSCRIPT/SUBSCRIPT/SMALL/FULLWIDTH/HALFWIDTH
+        if (canonical(bro) == brackstack[sp].b) {
+          // insert bracket pair into sorted list
+          int bracki = bracks;
+          while (bracki && brackpairs[bracki - 1].o > brackstack[sp].o) {
+            brackpairs[bracki] = brackpairs[bracki - 1];
+            bracki--;
+          }
+          brackpairs[bracki].o = brackstack[sp].o;
+          brackpairs[bracki].c = i;
+          //printf("N0 > bracket %d %d\n", brackstack[sp].o, i);
+          bracks++;
+          top = sp;
+          break;
+        }
+      }
+    }
+  }
+#define dont_debug_N0
+  // now handle the determined bracket pairs, assuming they are sorted
+  // • For each bracket-pair element in the list of pairs of text positions
+  for (int k = 0; k < bracks; k++) {
+    int o = brackpairs[k].o;
+    int c = brackpairs[k].c;
+#ifdef debug_N0
+    printf("N0 < bracket %d %d\n", o, c);
+#endif
+    bool otherstrong = false;
+    // a. Inspect the bidirectional types of the characters enclosed within the bracket pair.
+    for (int j = o + 1; j < c; j++) {
+      uchar type = types[j];
+      // b. If any strong type (either L or R) matching the embedding 
+      // direction is found, set the type for both brackets in the pair 
+      // to match the embedding direction.
+      if ((type == L && !(embeddingLevel & 1)) || (is_N0_R(type) && (embeddingLevel & 1))) {
+#ifdef debug_N0
+        printf("N0 strong emb %d\n", embeddingLevel);
+#endif
+        types[o] = types[c] = (embeddingLevel & 1) ? R : L;
+        otherstrong = false;
+        break;
+      }
+      // c. Otherwise, if there is a strong type it must be 
+      // opposite the embedding direction.
+      else if (type == L || is_N0_R(type)) {
+        otherstrong = true;
+      }
+    }
+    if (otherstrong) {
+      // c. Otherwise, ...
+      // Therefore, test for an established context with a preceding 
+      // strong type by checking backwards before the opening paired 
+      // bracket until the first strong type (L, R, or sos) is found.
+      int precedingLevel = embeddingLevel;
+      for (int j = o - 1; j >= 0; j--)
+        if (types[j] == L || is_N0_R(types[j])) {
+          precedingLevel = (embeddingLevel & ~1) + (types[j] != L);
+          //printf("N0 strong other %d preceding @%d\n", embeddingLevel, j);
+          break;
+        }
+#ifdef debug_N0
+      printf("N0 strong other %d preceding %d\n", embeddingLevel, precedingLevel);
+#endif
+      // 1. If the preceding strong type is also opposite the 
+      // embedding direction, context is established, so set the type for 
+      // both brackets in the pair to that direction.
+      if ((precedingLevel & 1) != (embeddingLevel & 1))
+        types[o] = types[c] = (precedingLevel & 1) ? R : L;
+      // 2. Otherwise set the type for both brackets in the pair to 
+      // the embedding direction.
+      else
+        types[o] = types[c] = (embeddingLevel & 1) ? R : L;
+    }
+    // d. Otherwise, there are no strong types within the bracket pair. 
+    // Therefore, do not set the type for that bracket pair.
+    // • Any number of characters that had original bidirectional 
+    // character type NSM prior to the application of W1 that immediately 
+    // follow a paired bracket which changed to L or R under N0 should 
+    // change to match the type of their preceding bracket.
+    if (types[c] == L || is_N0_R(types[c]))
+      for (int j = c + 1; j < count; j++) {
+        if (bidi_class_of(j) == NSM /*&& !skip[j]*/)
+          types[j] = types[c];
+        else
+          break;
+      }
+  }
+  trace_bidi("[N0]");
 
  /* Rule (N1)
   * N1. A sequence of NIs takes the direction of the surrounding
@@ -861,16 +1187,18 @@ do_bidi(bidi_char * line, int count)
     types are used at isolating run sequence boundaries.
   */
   if (count >= 2 && is_NI(types[0])) {
-    if ((types[1] == R) || (types[1] == EN) || (types[1] == AN))
+    if ((paragraphLevel & 1) &&
+        ((types[1] == R) || (types[1] == EN) || (types[1] == AN))
+       )
       types[0] = R;
-    else if (types[1] == L)
+    else if (!(paragraphLevel & 1) && types[1] == L)
       types[0] = L;
   }
-  for (i = 1; i < (count - 1); i++) {
+  for (i = 1; i < count - 1; i++) {
     if (is_NI(types[i])) {
       if (types[i - 1] == L) {
         j = i;
-        while (j < (count - 1) && is_NI(types[j])) {
+        while (j < count - 1 && is_NI(types[j])) {
           j++;
         }
         if (types[j] == L) {
@@ -884,7 +1212,7 @@ do_bidi(bidi_char * line, int count)
       else if ((types[i - 1] == R) || (types[i - 1] == EN) ||
                (types[i - 1] == AN)) {
         j = i;
-        while (j < (count - 1) && is_NI(types[j])) {
+        while (j < count - 1 && is_NI(types[j])) {
           j++;
         }
         if ((types[j] == R) || (types[j] == EN) || (types[j] == AN)) {
@@ -897,12 +1225,14 @@ do_bidi(bidi_char * line, int count)
     }
   }
   if (count >= 2 && is_NI(types[count - 1])) {
-    if (types[count - 2] == R || types[count - 2] == EN ||
-        types[count - 2] == AN)
+    if ((paragraphLevel & 1) &&
+        (types[count - 2] == R || types[count - 2] == EN || types[count - 2] == AN)
+       )
       types[count - 1] = R;
-    else if (types[count - 2] == L)
+    else if (!(paragraphLevel & 1) && types[count - 2] == L)
       types[count - 1] = L;
   }
+  trace_bidi("[N1]");
 
  /* Rule (N2)
   * N2. Any remaining NIs take the embedding direction.
@@ -915,33 +1245,29 @@ do_bidi(bidi_char * line, int count)
         types[i] = R;
     }
   }
+  trace_bidi("[N2]");
 
-  trace_bidi("<I1");
- /* Rule (I1)
+ /* Rules (I1) (I2)
   * I1. For all characters with an even (left-to-right) embedding
   * direction, those of type R go up one level and those of type AN or
   * EN go up two levels.
+  * I2. For all characters with an odd (right-to-left) embedding direction,
+  * those of type L, EN or AN go up one level.
   */
   for (i = 0; i < count; i++) {
     if ((levels[i] % 2) == 0) {
+      // (I1)
       if (types[i] == R)
         levels[i] += 1;
       else if (types[i] == AN || types[i] == EN)
         levels[i] += 2;
     }
-  }
-
- /* Rule (I2)
-  * I2. For all characters with an odd (right-to-left) embedding direction,
-  * those of type L, EN or AN go up one level.
-  */
-  for (i = 0; i < count; i++) {
-    if ((levels[i] % 2) == 1) {
+    else
+      // (I2)
       if (types[i] == L || types[i] == EN || types[i] == AN)
         levels[i] += 1;
-    }
   }
-  trace_bidi(">I2");
+  trace_bidi("[I1, I2]");
 
  /* Rule (L1)
   * L1. On each line, reset the embedding level of the following characters
@@ -957,19 +1283,31 @@ do_bidi(bidi_char * line, int count)
     N/A: Because a paragraph separator breaks lines, there will be at most 
     one per line, at the end of that line.
   */
+  // (4)
   j = count - 1;
-  while (j > 0 && (bidi_class_of(j) == WS)) {
+  while (j > 0 && (bidi_class_of(j) == WS || skip[j])) {
     j--;
   }
-  if (j < (count - 1)) {
+  if (j < count - 1) {
     for (j++; j < count; j++)
       levels[j] = paragraphLevel;
   }
+  // (3)
   for (i = 0; i < count; i++) {
     tempType = bidi_class_of(i);
+#ifdef check_emojiseq
+    if (line[i].emojilen) {
+      // raise emoji sequence to LTR level, as a rough approximation of
+      // HL3. Emulate explicit directional formatting characters.
+      for (j = i; j < i + line[i].emojilen && j < count && line[j].emojilen; j++) {
+        levels[j] = leastGreaterEven(levels[i]);
+      }
+    }
+    else
+#endif
     if (tempType == WS) {
       j = i;
-      while (j < count && (bidi_class_of(j) == WS)) {
+      while (j < count && (bidi_class_of(j) == WS || skip[j])) {
         j++;
       }
       if (j == count || bidi_class_of(j) == B || bidi_class_of(j) == S) {
@@ -983,6 +1321,10 @@ do_bidi(bidi_char * line, int count)
     }
   }
 
+#ifdef TEST_BIDI
+  memcpy(bidi_levels, levels, count);
+#endif
+
  /* Rule (L4)
   * L4. A character that possesses the mirrored property as specified by
   * Section 4.7, Mirrored, must be depicted by a mirrored glyph if the
@@ -991,7 +1333,7 @@ do_bidi(bidi_char * line, int count)
  /* Note: this is implemented before L2 for efficiency */
   for (i = 0; i < count; i++)
     if ((levels[i] % 2) == 1)
-      line[i].wc = mirror(line[i].wc);
+      line[i].wc = mirror(line[i].wc, box_mirror);
 
  /* Rule (L2)
   * L2. From the highest level found in the text to the lowest odd level on
@@ -1001,7 +1343,15 @@ do_bidi(bidi_char * line, int count)
   */
  /* we flip the character string and leave the level array */
   i = 0;
+#if __GNUC__ >= 13
+#pragma GCC diagnostic push
+// workaround wrong data flow analysis
+#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
+#endif
   tempType = levels[0];
+#if __GNUC__ >= 13
+#pragma GCC diagnostic pop
+#endif
   while (i < count) {
     if (levels[i] > tempType)
       tempType = levels[i];
@@ -1012,6 +1362,7 @@ do_bidi(bidi_char * line, int count)
     flip_runs(line, levels, tempType, count);
     tempType--;
   }
+  trace_bidi("[L1] [L4] [L2]");
 
  /* Rule (L3) NOT IMPLEMENTED
   * L3. Combining marks applied to a right-to-left base character will at
@@ -1023,5 +1374,5 @@ do_bidi(bidi_char * line, int count)
   // This is not relevant for mintty as the combining characters are kept 
   // hidden from this algorithm and are maintained transparently to it.
 
-  return R;
+  return resLevel;
 }
